
class UnrealPickup : ScriptObject
{
	kActor@ self;
	int nFrames;

	UnrealPickup( kActor@ a ) { @self = a; }

	void OnSpawn()
	{
		self.Definition().GetInt( "nFrames", nFrames );
	}

	void OnPostBeginLevel()
	{
		// if we appear to be outside the level, lift to original model's elevation
		if ( self.SectorIndex() < 0 || self.Origin().z - self.FloorHeight() <= -GAME_SCALE/4 )
		{
			float lift;
			if ( self.Definition().GetFloat( "lift", lift ) )
				self.RenderModel().Offset().z = lift * self.Scale().z;
		}
	}

	void OnTick()
	{
		if ( nFrames > 1 )
			self.ModelVariation() = self.GameTicks() % nFrames;
	}
}

//==============================================================================

class Enforcer : ScriptObject
{
	kActor@ self;
	Enforcer( kActor@ a ) { @self = a; }
	void OnSpawn() {}

	~Enforcer()
	{
		if ( self.GetTarget() !is null && !self.GetTarget().IsStale() )
			self.GetTarget().Remove();
	}

	void OnBeginLevel()
	{
		kActor@ a = ActorFactory.Spawn( "Enforcer_Generic",
			self.Origin().x, self.Origin().y, self.Origin().z, 0, self.SectorIndex() );
		self.SetTarget( a );
	}

	void OnTick()
	{
		// generic pickup was taken
		if ( self.GetTarget() is null || self.GetTarget().IsStale() )
		{
			int ammo;
			self.Definition().GetInt( "pickup.weapon.giveAmmo", ammo );
			Player.GiveWeapon( UT_Wep_Enforcer, ammo );
			SetAkimboEnforcer( true );
			self.MarkPersistentBit( false );
			self.Remove();
		}
		// player has 0 or 2 - no special logic needed - can just use normal weapon pickup
		else if ( !Player.HasWeapon( UT_Wep_Enforcer ) || HasAkimboEnforcer() )
		{
			self.Flags()             |=  AF_CANBETOUCHED;
			self.GetTarget().Flags() &= ~AF_CANBETOUCHED;
		}
		// player has 1 - need to allow picking up a second even if ammo is full - use generic pickup
		else
		{
			self.Flags()             &= ~AF_CANBETOUCHED;
			self.GetTarget().Flags() |=  AF_CANBETOUCHED;
		}
	}
}

class UnrealChrono : ScriptObject
{
	kActor@ self;
	UnrealChrono( kActor@ a ) { @self = a; }
	void OnSpawn() {}

	~UnrealChrono()
	{
		if ( self.GetTarget() !is null && !self.GetTarget().IsStale() )
			self.GetTarget().Remove();
	}

	void OnBeginLevel()
	{
		// already taken
		// could avoid this check by spawning in OnPostBeginLevel(), but spawning earlier allows UnrealPickups
		// to do their OnPostBeginLevel() elevation check without special handling
		if ( self.IsPersistentMarked() )
			return;
		kStr str;
		if ( !self.Definition().GetString( "pickupClass", str ) )
			return;
		kActor@ a = ActorFactory.Spawn( str, self.Origin().x, self.Origin().y, self.Origin().z, 0, self.SectorIndex() );
		if ( a is null ) return;
		a.Scale().Set( 0.35f, 0.35f, 0.35f );
		a.Flags() |= AF_CASTSHADOW;
		if ( a.InstanceOf("kexAmmoPickup") )
			a.Radius() = 25.6f;
		else
			a.Radius() = 40.96f;
		self.SetTarget( a );
	}

	void OnTick()
	{
		// hide the actual chrono piece (doing this in OnBeginLevel() sometimes doesn't work for first piece for some reason)
		self.Flags() |= AF_HIDDEN;
		if ( self.GetTarget() !is null && !self.GetTarget().IsStale() )
			return;

		int flagNum, flags;
		switch ( self.Type() )
		{
			case AT_PICKUP_CHRONOPIECE1: flagNum = 0; break;
			case AT_PICKUP_CHRONOPIECE2: flagNum = 1; break;
			case AT_PICKUP_CHRONOPIECE3: flagNum = 2; break;
			case AT_PICKUP_CHRONOPIECE4: flagNum = 3; break;
			case AT_PICKUP_CHRONOPIECE5: flagNum = 4; break;
			case AT_PICKUP_CHRONOPIECE6: flagNum = 5; break;
			case AT_PICKUP_CHRONOPIECE7: flagNum = 6; break;
			case AT_PICKUP_CHRONOPIECE8: flagNum = 7; break;
		}
		GameVariables.GetInt( "chronoPieceFlags", flags );
		GameVariables.SetValue( "chronoPieceFlags", ""+ (flags | (1 << flagNum)) );

		self.MarkPersistentBit( false );
		self.Remove();
	}
}

//==============================================================================

// change keg back to "+" for Turok+ title screen
class HealthKeg : UnrealPickup
{
	HealthKeg( kActor@ a ) { super( a ); }
	void OnPostBeginLevel()
	{
		if ( Game.GetCurrentMapID() == 42 )
		{
			self.RenderModel().SetModel( "models/health_ultra.bin", "" );
			nFrames = 0;
		}
		else
			UnrealPickup::OnPostBeginLevel();
	}
}

class UnrealArmorPickup : ScriptObject
{
	kActor@ self;
	UnrealArmorPickup( kActor@ a ) { @self = a; }
	void OnSpawn() {}

	// cap to 50 displayed armor after being picked up (and level end, which is redundant but fine)
	~UnrealArmorPickup()
	{
		if ( Player.Armor() > 150 ) Player.Armor() = 150;
	}

	// only allow picking up if displayed armor is < 50
	void OnTick()
	{
		if ( Player.Armor() <= 147 ) self.Flags() |=  AF_CANBETOUCHED;
		else                         self.Flags() &= ~AF_CANBETOUCHED;
	}
}
