
kActor@ Dummy1, Dummy2;

class UnrealProjectile : ScriptObject
{
	kActor@ self;
	int spawnTick;
	float timeScale = 1;
	bool bAmp;

	UnrealProjectile( kActor@ a )
	{
		@self = a;
		CheckAmp();
	}

	void Collision( kActor@ a, kVec3&in hitLoc, kVec3&in hitNormal, uint clipResult )
	{
	}

	//==================================

	void Destroy()
	{
		// don't block player while stale waiting to be purged
		self.Flags() &= ~AF_SOLID;
		// prevent sound from flying off
		self.Velocity() = Math::vecZero;
		self.Gravity() = 0;
		self.Remove();
	}

	int LifeTicks()
	{
		return self.GameTicks() > 0 ? PlayLoop.Ticks() - spawnTick : 0;
	}

	//==================================

	void OnSpawn()
	{
		spawnTick = PlayLoop.Ticks();
		// CF_COLLIDEFLOORS - prevents proj from jumping up slopes on collision?
		// CF_NOCEILINGADJUST results in slightly more consistent ceiling adjustment
		self.ClipFlags() = CF_CLIPEDGES | CF_DROPOFF | CF_ALLOWCRAWL | CF_COLLIDECORPSES | CF_NOCEILINGADJUST | CF_COLLIDEFLOORS;
		self.Rotation() = Player.Actor().Rotation();
		// make sure it doesn't freeze in mid air at long range
		self.Flags() |= AF_ALWAYSACTIVE;
	}

	void CheckAmp()
	{
		bAmp = UDamage();
	}

	void OnTick()
	{
		// sometimes we'll hit the floor and stop but OnCollide() won't be called
		if ( self.OnGround() )
			Collision( null, self.Origin(), kVec3(0,0,1), 1/*CRF_FLOOR*/ );

		float speed = self.Velocity().Unit();
		if ( speed > 0 )
			timeScale = self.Movement().Unit() / speed;
	}

	void OnCollide( kCModel@ cModel )
	{
		if ( cModel.ContactActor() is Player.Actor().CastToActor() )
			return;
		if ( cModel.ClipResult() & (1<<2) != 0 ) // wall
		{
			kVec3 o = self.PrevOrigin();
			o.z += self.Height() / 2;
			// make sure probes don't hit us
			float r = self.Radius();
			float h = self.Height();
			self.Radius() = 0;
			self.Height() = 0;
			// where we were
			Dummy1.SetPosition( o );
			// where we should be
			Dummy2.SetPosition( o += self.Velocity() * timeScale );
			// check if there's really an obstacle
			// if not, that means we really just hit the edge of an AAF_RESTRICTED sector
			if ( Dummy1.CanSee( Dummy2, CF_IGNOREBLOCKERS | CF_NOCLIPSTATICS | CF_NOCLIPACTORS ) )
			{
				o.z -= h/2;
				self.SetPosition( o );
				self.Radius() = r;
				self.Height() = h;
				return;
			}
			self.Radius() = r;
			self.Height() = h;
		}
		// ground gives (0,0,0) for some reason
		if ( cModel.ContactNormal().UnitSq() < 0.5f )
			cModel.ContactNormal().Set( 0, 0, 1 );
		Collision( cModel.ContactActor(), cModel.InterceptVector(), cModel.ContactNormal(), cModel.ClipResult() );
	}
}

//==============================================================================
//
//    Shock Proj
//
//==============================================================================

class ShockProj : UnrealProjectile
{
	ShockProj( kActor@ a )
	{
		super( a );
	}

	void OnSpawn()
	{
		UnrealProjectile::OnSpawn();

		self.Velocity() = kVec3(0,10,0) * self.Rotation();
		// makes shootable without being solid, but prevents OnCollide() from getting called with actors
//		self.Flags() |= AF_DEAD;
		//self.Flags() |= AF_SOLID; // moved to OnTick()
		// prevent corpses from nicking our large hit box
		self.ClipFlags() &= ~CF_COLLIDECORPSES;

		kActor@ hack = ActorFactory.Spawn( "ShockProjHack", 0,0,0, 0 );
		if ( hack is null ) return;
		hack.Flags() |= AF_ALWAYSACTIVE;
		hack.SetTarget( self );
	}

	void Explode( kVec3 hitNormal )
	{
//		kVec3 v = self.Origin() + hitNormal;// * self.Radius() / 2;
//		v.z += self.Height() / 2;
//		Game.SpawnFx( "UT/fx/ASMD/Impact.kfx", self, v, ToQuat(hitNormal) );
//		Game.SpawnFx( "UT/fx/ASMD/Impact.kfx", self, self.PrevOrigin() + kVec3(0,0,self.Height()/2), ToQuat(hitNormal) );
		self.SpawnFx( "UT/fx/ASMD/Impact.kfx", kVec3( 0, 0, self.Height() / 2 ), ToQuat(hitNormal) );
		self.PlaySound( "U1/sounds/ASMD/Impact.ksnd" );
		Destroy();
	}

	void OnDamage( kActor@ instigator, kDictMem@ damageDef, const int damage )
	{
		// only trigger if shot by ASMD beam
		bool bASMD;
		if ( damageDef is null || !damageDef.GetBool( "bASMD", bASMD ) || !bASMD )
			return;

		if ( Player.CurrentWeapon() == UT_Wep_ASMD )
			Player.ConsumeAmmo( 2 );

		kVec3 v( 0, 0, self.Height() / 2 );
		self.SpawnFx( "UT/fx/ASMD/Combo_Explosion.kfx", v );
		v += self.Origin();
		ActorFactory.Spawn( "ShockCombo", v.x, v.y, v.z, Player.Actor().Yaw(), self.SectorIndex() );
		self.PlaySound( "U1/sounds/ASMD/Combo.ksnd" );

		// hurt player with reduced radius
		float dist = Player.Actor().DistanceToPoint( self.Origin() + kVec3( 0, 0, self.Height() / 2 ) );
		dist = Math::Sqrt( dist ) - Player.Actor().Radius();
		if ( dist < 0 ) dist = 0;
		float pct = dist / (30*GAME_SCALE);
		pct = 1 - pct;//*pct;
		int dam = int( (bAmp ? 375 : 125) / 4.0f * pct );
		if ( dam > 0 )
			Player.Actor().InflictGenericDamage( Player.Actor().CastToActor(), dam );

		Destroy();
	}

	void OnTick()
	{
		if ( self.InWater() )
		{
			self.SpawnFx( "UT/fx/ASMD/Ball_Impact_Water.kfx", Math::vecZero );
			self.PlaySound( "sounds/shaders/water_splash_4.ksnd" );
			Destroy();
			return;
		}

		UnrealProjectile::OnTick();

		// 10s life span
		int elapsed = LifeTicks();
		if ( elapsed >= 10*60 )
			Destroy();
		// only become solid after alt fire anim ends, to avoid blocking player until we've moved out of the way
		else if ( elapsed == 30 )
			self.Flags() |= AF_SOLID;	}

	void Collision( kActor@ a, kVec3&in hitLoc, kVec3&in hitNormal, uint clipResult )
	{
//		if ( a !is null )
//			a.InflictGenericDamage( Player.Actor().CastToActor(), 21 );
		self.InteractActorsAtPosition( self.Origin(), "SplashDamage", bAmp ? 63 : 21, 10*GAME_SCALE );
		Explode( hitNormal );
	}

	void SplashDamage( kActor@ a, const float damage, const float radius, const float f3, const float f4 )
	{
		// direct hit, just do full damage
		if ( a is CModel.ContactActor() )
		{
			a.InflictGenericDamage( Player.Actor().CastToActor(), int(damage) );
			return;
		}

		float dam = damage;
		float rad = radius;
		if ( a is Player.Actor().CastToActor() )
			dam /= 4;

		kVec3 dif = a.Origin() - self.Origin();

		dif.x = ( dif * kVec3(1,1,0) ).Unit() - a.Radius() - self.Radius();
		if ( dif.x < 0 ) dif.x = 0;
		dif.y = 0;

		if ( dif.z < 0 )
		{
			dif.z += a.Height();
			if ( dif.z > 0 ) dif.z = 0;
		}
		else if ( dif.z > 0 )
		{
			dif.z -= self.Height();
			if ( dif.z < 0 ) dif.z = 0;
		}

		float f = dif.Unit();
		f = 1 - f / rad;
		int i = int( f * dam );
		if ( i <= 0 ) return;

		// can't see into bridge sectors, but can see out of bridge sectors
		// so require LOS to be blocked both ways to reject
		if ( !self.CanSee( a, CF_IGNOREBLOCKERS ) && !a.CanSee( self, CF_IGNOREBLOCKERS ) )
			return;

		a.InflictGenericDamage( Player.Actor().CastToActor(), i );
	}
}

// spawning proj's sprite in ShockProj::Ontick() causes it to be updated twice per tick most of the time if Turok+ isn't installed for some fucking reason
// so spawn it from a different actor's tick, to prevent actual projectile from moving twice as fast
// unfortunately, spawning the particle here still often causes this actor to be updated twice per tick, which means it spawns two instances of the sprite
// adding a check for this prevents the doubling, but also makes only the second particle in the .kfx visible much of the time, because this engine is shonky garbage
// adding an invisible dummy particle to the top seems to work around that (lifeTime must be > 0)
// the Quake 2 BFG projectile doesn't have the problem -- probably some weird edge case deep in the engine's update logic getting hit or dodged by chance
class ShockProjHack : ScriptObject
{
	kActor@ self;
	int lastTick = -1;
	ShockProjHack( kActor@ a )
	{
		@self = a;
	}
	void OnSpawn() {}
	void OnTick()
	{
		kActor@ proj = self.GetTarget();
		if ( proj is null || proj.IsStale() )
		{
			self.Remove();
			return;
		}

		if ( lastTick == Player.Actor().GameTicks() ) return;
		lastTick = Player.Actor().GameTicks();

		if ( proj.ScriptObject() is null ) return;
		ShockProj@ script = cast<ShockProj@>( proj.ScriptObject().obj );
		if ( script is null ) return;

		kStr str = "UT/fx/ASMD/Ball_";
		int frame = int( script.LifeTicks() / 1.5f );
		if ( frame < 4 ) str += "Spawn_";
		str = str + ( frame % 4 ) + ".kfx";
		proj.SpawnFx( str, kVec3( 0, 0, proj.Height() / 2 ) );
	}
}

class ShockCombo : ScriptObject
{
	int lastScaledTick = -1, scaledTickCount = 0;
	float timeScale = 1;
	float LifeSpan = 1.25f;
	kActor@ self;

	ShockCombo( kActor@ a )
	{
		@self = a;
	}
	void OnSpawn()
	{
		// make sure it doesn't freeze at long range
		self.Flags() |= AF_ALWAYSACTIVE;
	}

	void OnTick()
	{
		if ( lastScaledTick == PlayLoop.Ticks() )
			++scaledTickCount;
		else
		{
			timeScale = 1.0f / (1.0f + scaledTickCount);
			scaledTickCount = 0;
			lastScaledTick = PlayLoop.Ticks();
		}

		LifeSpan -= GAME_DELTA_TIME * timeScale;
		if ( LifeSpan <= 0 )
		{
			self.Remove();
			return;
		}
		float f = LifeSpan / 1.25f;
//		self.Scale().x = self.Scale().y = self.Scale().z = 0.7f / (f + 0.05f);
		self.Scale().x = self.Scale().y = self.Scale().z = 1.5f / (f + 0.15f);
		self.ModelVariation() = int( (1-f) * 10 );
	}
}

//==============================================================================
//
//    Bio Gel
//
//==============================================================================

array<kActor@> GESlist;

class BioGel : UnrealProjectile
{
	float LifeSpan = 3;
	int charge, nExplosionBlobs = 9;
	int lastScaledTick = 0;
	kVec3 surfaceNormal( 0, 0, 1 );
	bool bSubBlob, bSplashed = false;

	BioGel( kActor@ a )
	{
		super( a );
		GESlist.insertLast( a );
	}

	~BioGel()
	{
		int i = GESlist.findByRef( self );
		if ( i < 0 ) Sys.Warning( "couldn't find bio gel in list" );
		else         GESlist.removeAt( i );
	}

	void Destroy()
	{
		// try preventing Explode() from touching stale scripts objects?
//		LifeSpan = 0;
//		self.Flags() |= AF_DISABLED;

		self.Flags() &= ~AF_DEAD; // prevent other blobs from hitting us while stale waiting to be purged
		UnrealProjectile::Destroy();
	}

	// skip this and handle with Setup(), so sub blobs will match parent and won't cause amp sound
	void CheckAmp()
	{
	}

	void Setup( int i, kVec3&in dir, bool bAmplified, bool bSub = false )
	{
		charge = i;
//		float f = (charge < 0 ? 0 : charge) / 9.0f;
//		self.Velocity() = kVec3( 0, 840 * GAME_DELTA_TIME, 0 ) * self.Rotation();
//		self.Velocity() = kVec3( 0, 950 * (1 - f*0.25f) * GAME_DELTA_TIME, 0 ) * Player.Actor().Rotation();
//		self.Velocity() = kVec3( -0.0095f, 1, 0 ).Normalize() * Player.Actor().Rotation();
//		self.Velocity() *= 950 * GAME_DELTA_TIME;// * (1 - f*0.15f);
		self.Velocity() = dir * 950 * GAME_DELTA_TIME;// * (1 - f*0.15f);
		self.Gravity() = 0.25f;
		self.Velocity().z += 8 * self.Gravity();// * (1 + f*0.3f);
		self.RenderModel().Offset().z = self.Height() / 2;
		bAmp = bAmplified;
		bSubBlob = bSub;
	}

	void Splash()
	{
		self.PlaySound( "sounds/shaders/water_splash_3.ksnd" );
		Game.SpawnFx( "U/fx/Water_Splash_Small.kfx", self.Origin(), self.SectorIndex() );
		bSplashed = true;
	}

	void OnTick()
	{
		UnrealProjectile::OnTick();
		if ( LifeTicks() * GAME_DELTA_TIME >= LifeSpan )
		{
			Explode( surfaceNormal );
			return;
		}

		float scale = self.Scale().x / 0.35f;
		scale += timeScale / 2;
		float max = 1 + 1.5f * charge / 9.0f;
		if ( scale > max ) scale = max;
		self.Scale().x = self.Scale().y = self.Scale().z = scale * 0.35f;

		// still flying
		if ( self.Gravity() != 0 )
		{
			if ( self.InWater() )
			{
				// timeScale is calculated based on our velocity, which gets close to 0, causing spikes
				// so just apply physics once per scaled tick instead
				if ( PlayLoop.Ticks() != lastScaledTick )
				{
					self.Velocity() *= 0.9f;
					self.Velocity().z += 0.5f;
				}
				if ( !bSplashed )
					Splash();
			}
			lastScaledTick = PlayLoop.Ticks();
			return;
		}

		// check if we were stepped on
		self.InteractActorsAtPosition( self.Origin(), "Touched" );
		if ( self.GetTarget() !is null )
		{
			self.SetTarget( null ); // don't treat this as a direct hit
			Explode( surfaceNormal );
			return;
		}

		lastScaledTick = PlayLoop.Ticks();

		// spawn sub blobs if charged
		if ( charge < 1 ) return;
		--charge;
		kVec3 dir = surfaceNormal;
		dir.Randomize( 0.5f );
		kVec3 v = self.Origin() + dir * ( self.Height() + self.Radius() );
		kActor@ a = ActorFactory.Spawn( "BioGel", v.x, v.y, v.z, dir.ToYaw(), self.SectorIndex() );
		if ( a is null ) return;
		BioGel@ g = cast<BioGel@>( a.ScriptObject().obj );
		if ( g !is null )
			g.Setup( 0, dir * 0.35f, bAmp, true );
	}

	// exploding in here tends to crash the game, so just note the touch and react in OnTick()
	void Touched( kActor@ a, const float f1, const float f2, const float f3, const float f4 )
	{
		float r = self.Radius(), h = self.Height();
		if ( a !is Player.Actor().CastToActor() && a.Flags() & AF_STATIONARY == 0 )
		{
			r *= 2;
			h *= 1.5f;
		}

		if ( a.Origin().z > self.Origin().z + h / 2
		  || a.Origin().z + a.Height() / 2 < self.Origin().z )
				return;

		kVec3 dif = ( a.Origin() - self.Origin() ) * kVec3( 1, 1, 0 );
		if ( dif.UnitSq() <= r * r + a.Radius() * a.Radius() )
			self.SetTarget( a );
	}

	// SetPosition() occassionally crashes the game here
	void Collision( kActor@ a, kVec3&in hitLoc, kVec3&in hitNormal, uint clipResult )
	{
		// hit actor
		if ( a !is null )
		{
			// prevent sub blobs from hitting each other
			if ( bSubBlob && a.ScriptObject() !is null && cast<BioGel@>( a.ScriptObject().obj ) !is null )
				return;
//			self.SetPosition( hitLoc );
			self.SetTarget( CModel.ContactActor() ); // make sure what we hit gets full damage
			Explode( hitNormal );
		}
		// hit level
		else if ( self.Gravity() != 0 )
		{
			if ( !bSplashed && (self.OnGround() && World.GetAreaFlags(self.AreaID()) & (AAF_WATER|AAF_SHALLOWWATER) == (AAF_WATER|AAF_SHALLOWWATER)) )
				Splash();

//			self.SetPosition( hitLoc );
			self.Velocity() = Math::vecZero;
			self.Gravity() = 0;
			// only reset life time if not chain reacting
			if ( LifeSpan > 0.5f )
			{
				spawnTick = PlayLoop.Ticks();
				LifeSpan = 3.8f;
			}
			surfaceNormal = hitNormal;
			self.AnimState().Set( anim_weaponFire, 4, ANF_LOOP );
			// allow other blobs to hit us, and ride elevators
			self.Flags() |= AF_DEAD | AF_SNAPTOFLOOR;
			self.PlaySound( "U1/sounds/Bio/Splat.ksnd" );
//			Game.SpawnFx( "UT/fx/Bio/Splat.kfx", self, self.Origin() + hitNormal * self.Radius(), ToQuat( hitNormal ) );
			Game.SpawnFx( "UT/fx/Bio/Splat.kfx", self, hitLoc, ToQuat( hitNormal ) );
		}
	}

	void Explode( kVec3&in hitNormal )
	{
		kVec3 v = self.Origin() + hitNormal * self.Radius() + kVec3( 0, 0, self.Height()/2 );
		kQuat q = ToQuat( hitNormal );
		Game.SpawnFx( "UT/fx/Bio/Explosion.kfx", self, v, q );
		while ( nExplosionBlobs-- > 0 )
			Game.SpawnFx( "U1/fx/Bio/Explosion_Blob.kfx", self, v, q );
		self.PlaySound( "U1/sounds/Bio/Explode.ksnd" );
		float f = charge / 9.0f;
		// 21 is just enough to 1-shot poachers on hard; 215 is just enough for pur-lin
//		int dam = int( (bAmp ? 63 : 21) * (1 + 10 * f) );
		int dam = int( Math::Lerp( 21, 215, f ) );
		if ( bAmp ) dam *= 3;
		self.InteractActorsAtPosition( self.Origin(), "SplashDamage", dam, (15 + 5*f) * GAME_SCALE );

		kVec3 center = self.Origin();
		center.z += self.Height() / 2;
		float sqrRad = 15 * GAME_SCALE;
		sqrRad *= sqrRad;
		for ( uint i=0; i<GESlist.length(); ++i )
		{
			kActor@ a = GESlist[i];
			if ( a is null || a.IsStale() || a.Flags() & AF_DISABLED != 0 || a.ScriptObject() is null || a.ScriptObject().obj is null )
				continue;
			BioGel@ g = cast<BioGel@>( a.ScriptObject().obj );
			// don't reset timer if already chain reacting
			if ( g is null || g.LifeSpan < 0.5f || a.DistanceToPoint(center) > sqrRad )
				continue;
			// can't see into bridge sectors, but can see out of bridge sectors
			// so require LOS to be blocked both ways to reject
			if ( !self.CanSee( a, CF_IGNOREBLOCKERS ) && !a.CanSee( self, CF_IGNOREBLOCKERS ) )
				continue;
			g.spawnTick = PlayLoop.Ticks();
//			g.LifeSpan = 0.1f + Math::RandFloat() * 0.2f;
			g.LifeSpan = 0.05f + Math::RandFloat() * 0.3f;
			g.nExplosionBlobs = 6;
		}

		Destroy();
	}

	void SplashDamage( kActor@ a, const float damage, const float radius, const float f3, const float f4 )
	{
		// direct hit, just do full damage
		if ( a is self.GetTarget() )
		{
			a.InflictGenericDamage( Player.Actor().CastToActor(), int(damage) );
			return;
		}

		float dam = damage;
		float rad = radius;
		if ( a is Player.Actor().CastToActor() )
			dam /= 4;

//		float f = a.DistanceToPoint( self.Origin() + kVec3(0,0,self.Height()/2) );
//		f = Math::Sqrt( f ) - a.Radius();

		kVec3 dif = a.Origin() - self.Origin();

		dif.x = ( dif * kVec3(1,1,0) ).Unit() - a.Radius();// - self.Radius();
		if ( dif.x < 0 ) dif.x = 0;
		dif.y = 0;

		if ( dif.z < 0 )
		{
			dif.z += a.Height();
			if ( dif.z > 0 ) dif.z = 0;
		}
		else if ( dif.z > 0 )
		{
			dif.z -= self.Height();
			if ( dif.z < 0 ) dif.z = 0;
		}

		float f = dif.Unit();
		if ( f >= rad ) return;

		// can't see into bridge sectors, but can see out of bridge sectors
		// so require LOS to be blocked both ways to reject
		if ( !self.CanSee( a, CF_IGNOREBLOCKERS ) && !a.CanSee( self, CF_IGNOREBLOCKERS ) )
			return;

		// this only works if solid and not dead
/*		if ( a.ScriptObject() !is null )
		{
			BioGel@ g = cast<BioGel@>( a.ScriptObject().obj );
			if ( g !is null )
			{
				g.spawnTick = PlayLoop.Ticks();
				g.LifeSpan = 0.1f + Math::RandFloat() * 0.2;
				return;
			}
		}
*/
		f = 1 - f / rad;
		int i = int( f * dam );
		if ( i > 0 )
			a.InflictGenericDamage( Player.Actor().CastToActor(), i );
	}
}

//==============================================================================
//
//    Redeemer Missile
//
//==============================================================================

class RedeemerMissile : UnrealProjectile
{
	int lastTick = -1;

	RedeemerMissile( kActor@ a )
	{
		super( a );
	}

	void OnSpawn()
	{
		UnrealProjectile::OnSpawn();

//		bQuad = QuadDamage();
		self.Velocity() = kVec3(0,700,0) * self.Rotation() * GAME_DELTA_TIME;
//		lastTick = PlayLoop.Ticks() + 1; // don't spawn smoke in the player's face on first tick
		self.Scale().Set( 0,0,0 );
	}

	void Explode( kVec3 hitNormal )
	{
		kVec3 v = self.Origin() + hitNormal * 50;
		kStr fx = "UT/fx/Redeemer/Explosion";
		if ( bAmp ) fx += "_Amp";
		Game.SpawnFx( fx + ".kfx", self, v, ToQuat(hitNormal) );
		kActor@ a = ActorFactory.Spawn( "RedeemerExplosion", v.x, v.y, v.z, hitNormal.ToYaw(), self.SectorIndex() );
		// source sounds from explosion actor, so they'll source from the same point as second explosion sprite
		if ( a !is null )
		{
			a.PlaySound( "UT/sounds/Redeemer/Impact.ksnd" );
			a.PlaySound( "UT/sounds/Redeemer/Explode.ksnd" );
		}

		kVec3 center = Player.Actor().Origin();
		center.z += Player.Actor().Height()/2;
		float power = 1 - ( self.Origin() - center ).Unit() / (250 * GAME_SCALE);
		if ( power > 0 )
		{
			power = Math::Pow( power, 1.5f );
			TurokPlayer@ p = cast<TurokPlayer@>( Player.Actor().ScriptObject().obj );
			p.DoViewShake( 0, 0.1f * power, 2 );
		}

		Destroy();
	}

	void OnTick()
	{
		if ( self.InWater() )
		{
			Explode( kVec3(0,0,1) );
			return;
		}

		UnrealProjectile::OnTick();

		// 10s life span
		int elapsed = LifeTicks();
		if ( elapsed >= 10*60 )
		{
			Destroy();
			return;
		}

		self.Rotation() = ToQuat( self.Velocity() );
		self.PlaySound( "UT/sounds/Redeemer/Fly.ksnd" );
		// scale up from nothing to mask the fact that the rocket spawns right in your face
		// don't scale this with game speed, so it'll always be imperceptible
		if ( self.Scale().y < 0.35f )
		{
			self.Scale() += kVec3(0.35f,0.35f,0.35f) * GAME_DELTA_TIME * 20;
			if ( self.Scale().y > 0.35f )
				self.Scale().Set( 0.35f, 0.35f, 0.35f );
		}
		// don't spawn trail until we're out of the player's face
		if ( elapsed < 6 ) return;

		if ( PlayLoop.Ticks() == lastTick ) return;
		lastTick = PlayLoop.Ticks();

		kVec3 v = self.Velocity();
		self.Velocity() += v.Normalize() * 100 * GAME_DELTA_TIME * GAME_DELTA_TIME;

		v = self.Origin() + kVec3( 0, -175, 0 ) * self.Scale() * self.Rotation();
		Game.SpawnFx( "UT/fx/Rocket/Rocket_Trail.kfx", self, v, self.Rotation() );
		Game.SpawnFx( "UT/fx/Redeemer/Trail.kfx", self, v, self.Rotation() );
		Game.SpawnFx( "UT/fx/Redeemer/Trail.kfx", self, v - self.Velocity() * 0.5f, self.Rotation() );
	}

	void Collision( kActor@ a, kVec3&in hitLoc, kVec3&in hitNormal, uint clipResult )
	{
		Explode( hitNormal );
	}
}

class RedeemerMissile_Guided : RedeemerMissile
{
	kAngle oldPitch, oldYaw, targetPitch, targetYaw;

	RedeemerMissile_Guided( kActor@ a )
	{
		super( a );
	}
	~RedeemerMissile_Guided()
	{
		Camera.StartCinematic( CMF_NO_LETTERBOX | CMF_SHOW_HIDDEN_OBJECTS | CMF_INITIAL_FADEIN );
		Player.Actor().Flags() |= AF_NODRAW;
		Player.Actor().Pitch() = oldPitch;
	}

	void OnSpawn()
	{
		UnrealProjectile::OnSpawn();
//		bQuad = QuadDamage();
		self.Velocity() = kVec3(0,550,0) * self.Rotation() * GAME_DELTA_TIME;

		Camera.StartCinematic( CMF_NO_LETTERBOX | CMF_SHOW_HIDDEN_OBJECTS | CMF_NO_INITIAL_FADEOUT );
		// at least one of these allows camera position to be updated in OnTick() after key acquisition cutscene
		Camera.ClearLookAtActor();
		Camera.ClearFinalView();
		Camera.ClearViewTracks();
		kStr str;
		Camera.fov = ( Sys.GetCvarValue( "r_fov", str ) ? str.Atof() : 74 );

		Player.Actor().AnimState().Set( 2801, 4, 0 );
		Player.Actor().AnimState().Pause();
		Player.Actor().Flags() &= ~AF_NODRAW;
		// these crash the game
//		targetPitch = oldPitch = Player.Actor().Pitch();
//		targetYaw   = oldYaw   = Player.Actor().Yaw();
		// but this is fine
		oldPitch = Player.Actor().Pitch();
		oldYaw   = Player.Actor().Yaw();
		targetPitch = oldPitch;
		targetYaw   = oldYaw;
		Player.Actor().Pitch() = 0;
	}

	void OnTick()
	{
		if ( self.InWater() )
		{
			Explode( kVec3(0,0,1) );
			return;
		}

		UnrealProjectile::OnTick();

		// 60s life span
		int elapsed = LifeTicks();
		if ( elapsed >= 60*60 )
		{
			Destroy();
			return;
		}

		targetPitch += Player.Actor().Pitch();
		targetYaw   += Player.Actor().Yaw() - oldYaw;
		Player.Actor().Pitch() = 0;
		Player.Actor().Yaw()   = oldYaw;

		kQuat q = kQuat( targetYaw, 0,0,1 ) * kQuat( targetPitch, 1,0,0 );
//		self.Velocity() = kVec3(0,550,0) * q * GAME_DELTA_TIME;
		self.Velocity() += kVec3( 0, 1500 * GAME_DELTA_TIME * GAME_DELTA_TIME, 0 ) * q;
		self.Velocity() = self.Velocity().Normalize() * 550 * GAME_DELTA_TIME;
		self.Rotation() = ToQuat( self.Velocity() );
		Camera.origin = self.Origin();
		Camera.pitch  = -self.Velocity().ToPitch();
		Camera.yaw    = self.Velocity().ToYaw();
		CameraView();
		self.PlaySound( "UT/sounds/Redeemer/Fly.ksnd" );

		if ( PlayLoop.Ticks() == lastTick ) return;
		lastTick = PlayLoop.Ticks();

		kVec3 v = self.Origin() + kVec3( 0, -175, 0 ) * self.Scale() * self.Rotation();
		Game.SpawnFx( "UT/fx/Rocket/Rocket_Trail.kfx", self, v, self.Rotation() );
//		Game.SpawnFx( "UT/fx/Redeemer/Trail.kfx", self, v, self.Rotation() );
//		Game.SpawnFx( "UT/fx/Redeemer/Trail.kfx", self, v - self.Velocity() * 0.5f, self.Rotation() );
	}

	void CameraView()
	{
		float distScale = 1.0f / Math::Tan( Math::Deg2Rad( Camera.fov ) / 2.0f );
		kQuat rot = kQuat( Camera.yaw  , 0,0,1 ) *
		            kQuat( Camera.pitch, 1,0,0 ) *
		            kQuat( Camera.roll,  0,1,0 );
		kVec3 loc = Camera.origin + kVec3( 0, distScale, 0 ) * rot;
		Game.SpawnFx( "UT/fx/Redeemer/Reticle.kfx", self, loc, rot );
	}
/*
	// allow collision with player
	void OnCollide( kCModel@ cModel )
	{
		if ( cModel.ContactActor() is Player.Actor().CastToActor() )
			Collision( cModel.ContactActor(), cModel.InterceptVector(), cModel.ContactNormal(), cModel.ClipResult() );
		else
			UnrealProjectile::OnCollide( cModel );
	}*/
}

class RedeemerExplosion : ScriptObject
{
	int lastScaledTick = -1, scaledTickCount = 0;
	float timeScale = 1;
	kActor@ self;

	RedeemerExplosion( kActor@ a )
	{
		@self = a;
	}
	void OnSpawn()
	{
		// make sure it doesn't freeze at long range
		self.Flags() |= AF_ALWAYSACTIVE;
	}

	void OnTick()
	{
		if ( lastScaledTick == PlayLoop.Ticks() )
			++scaledTickCount;
		else
		{
			timeScale = 1.0f / (1.0f + scaledTickCount);
			scaledTickCount = 0;
			lastScaledTick = PlayLoop.Ticks();
		}
		float min = 0;//30;
		float max = 100;
		self.Scale().x = self.Scale().y = self.Scale().z += (max - min) / 1.25f * GAME_DELTA_TIME * timeScale;
//		max *= 1.1f;
		if ( self.Scale().x >= max )
		{
			self.Remove();
			return;
		}
		float f = (self.Scale().x - min) / (max-min);
		self.ModelVariation() = int( f*f * 10 );
	}
}
